import os
import paramiko
import yaml

from si_tests import settings
from si_tests.deployments.utils import commons, file_utils, wait_utils
from si_tests.deployments.utils.namespace import NAMESPACE
from . import base


class RefappHeat(base.RefappBase):
    def __init__(self, client, kubeconfig=None):
        super().__init__(client, kubeconfig)
        self.stack_name = "refapp-test"

    @property
    def url(self):
        if self._url is None:
            stack_outputs = self.ostcm.stack.show([self.stack_name, "-c", "outputs"])["outputs"]
            self._url = list(filter(lambda x: x["output_key"] == "app_url", stack_outputs))[0]["output_value"]
        return self._url

    @property
    def _is_active(self):
        """Whether heat stack is active or not"""

        return self.ostcm.stack.exists(self.stack_name)

    def deploy(self):
        if not self._is_active:
            source = file_utils.join(commons.REFAPP_DIR, "heat-templates")
            if not os.path.exists(source):
                commons.LOG.warning("Refapp source does not exist %s", source)

            client_name = self.ostcm.client.name
            destination = f"{NAMESPACE.openstack}/{client_name}:/tmp/heat-templates"

            commons.LOG.info("Copy heat templates")
            self.kubectl.cp(source, destination)

            commons.LOG.info("Create heat stack")
            stack_parameters = [
                "--parameter",
                f"database_admin_password={self.db_passwords['admin']}",
                "--parameter",
                f"app_database_password={self.db_passwords['app']}",
            ]
            if settings.OPENSTACK_REFAPP_IMAGE:
                stack_parameters.extend(["--parameter", f"app_docker_image={settings.OPENSTACK_REFAPP_IMAGE}"])

            self.ostcm.stack.create([self.stack_name, "-t", "/tmp/heat-templates/top.yaml", *stack_parameters])

            commons.LOG.info("Waiting for heat stack to create ...")
            wait = wait_utils.Waiter(self.ostcm.os_manager, 3600)

            def f():
                status = self.ostcm.stack.show([self.stack_name])["stack_status"]
                if status == "CREATE_FAILED":
                    resource_list = self.ostcm.stack.resource_list(
                        ['-n', '10', self.stack_name])
                    event_list = self.ostcm.stack.event_list(
                        ['--nested-depth', '10', self.stack_name])
                    commons.LOG.error("Resource info: %s\nHeat stack event list: %s",
                                      resource_list, event_list)
                    raise Exception("Failed to create stack")
                return status == "CREATE_COMPLETE"

            wait.wait(f)

        commons.LOG.info("Refapp is successfully deployed")

    def delete(self):
        commons.LOG.info("Delete RefApp stack")
        self.ostcm.stack.delete([self.stack_name])

        wait = wait_utils.Waiter(self.ostcm.os_manager, 3600)

        def wait_refapp_stack_delete():
            return not self.ostcm.stack.exists(self.stack_name)

        commons.LOG.info("Wait for RefApp heat stack to delete")
        wait.wait(wait_refapp_stack_delete)

        commons.LOG.info("Refapp heat stack successfully deleted")

    def collect_logs(self, proxy_ip=None, dst_dir=settings.ARTIFACTS_DIR):
        """KUBECONFIG is required"""

        proxy_connection = None
        if proxy_ip:
            proxy_connection = commons.get_ssh_client(proxy_ip)
        elif not settings.get_var_as_bool("RUN_ON_REMOTE", False):
            proxy_ip = settings.SEED_STANDALONE_EXTERNAL_IP
            if proxy_ip:
                proxy_connection = commons.get_ssh_client(
                    ip=proxy_ip,
                    auth=commons.get_ssh_auth(
                        username=settings.SEED_SSH_LOGIN, key_filename=settings.SEED_SSH_PRIV_KEY_FILE
                    ),
                )
            else:
                proxy_ip = settings.PHYS_SEED_FIXED_IP
                if proxy_ip:
                    proxy_connection = commons.get_ssh_client(proxy_ip)
                else:
                    proxy_ip = self.ostcm.os_manager.api._K8sCluster__api_client.configuration.host
                    if proxy_ip:
                        proxy_ip = proxy_ip.split(":")[1][2:]
                        proxy_connection = commons.get_ssh_client(proxy_ip)
                    else:
                        commons.LOG.error("No proxy IP for access to refapp stack")
                        return

        stack_info = self.ostcm.stack.show([self.stack_name])
        if len(stack_info) > 1:
            vm_list = self.ostcm.stack.resource_list(
                [
                    "-n",
                    "10",
                    self.stack_name,
                    "-c",
                    "physical_resource_id",
                    "--filter",
                    "type=OS::Nova::Server",
                    "-f",
                    "value",
                ]
            ).split()
            if vm_list:
                commons.LOG.info("Collect logs from RefApp stack")
                work_dir = os.path.join(dst_dir, self.stack_name + ".log")
                os.mkdir(work_dir)
                instances_dir = os.path.join(work_dir, "instances")
                os.mkdir(instances_dir)
                with open(os.path.join(work_dir, "stack_info.yaml"), "w") as f:
                    yaml.safe_dump(stack_info, f)
                lb_list = self.ostcm.stack.resource_list(
                    [
                        "-n",
                        "10",
                        self.stack_name,
                        "-c",
                        "physical_resource_id",
                        "--filter",
                        "type=OS::Octavia::LoadBalancer",
                        "-f",
                        "value",
                    ]
                ).split()
                if lb_list:
                    current_dir = os.path.join(work_dir, "loadbalancer")
                    os.mkdir(current_dir)
                    for lb_item in lb_list:
                        lb_info = self.ostcm.loadbalancer.show([lb_item])
                        with open(os.path.join(current_dir, lb_info["name"] + ".yaml"), "w") as f:
                            yaml.safe_dump(lb_info, f)
                        amphora_list = self.ostcm.loadbalancer.amphora_list(
                            ["--loadbalancer", lb_item, "-c", "id", "-f", "value"]
                        ).split()
                        if amphora_list:
                            for amphora_item in amphora_list:
                                amphora_info = self.ostcm.loadbalancer.amphora_show([amphora_item])
                                with open(os.path.join(current_dir, "amphora-" + amphora_item + ".yaml"), "w") as f:
                                    yaml.safe_dump(amphora_info, f)
                                if amphora_info["compute_id"] is not None:
                                    vm_list.append(amphora_info["compute_id"])
                for instance_id in vm_list:
                    instance_info = self.ostcm.server.show([instance_id])
                    if instance_info and isinstance(instance_info, dict):
                        current_dir = os.path.join(instances_dir, instance_info["name"])
                        os.mkdir(current_dir)
                        with open(os.path.join(current_dir, "instance_info.yaml"), "w") as f:
                            yaml.safe_dump(instance_info, f)
                        try:
                            instance_ip_list = self.ostcm.server_get_floating_ips(instance_id)
                            if instance_ip_list:
                                if proxy_connection:
                                    ssh_connect = proxy_connection.proxy_to(
                                        host=instance_ip_list[0], auth=commons.get_ssh_auth()
                                    )
                                else:
                                    ssh_connect = commons.get_ssh_client(instance_ip_list[0])
                                self.ostcm.server_collect_logs(
                                    ssh_connect=ssh_connect, arch_path=current_dir, docker_logs=True
                                )
                            else:
                                self.ostcm.server_collect_console(instance_id, arch_path=current_dir)
                        except paramiko.ssh_exception.SSHException:
                            self.ostcm.server_collect_console(instance_id, arch_path=current_dir)
        else:
            commons.LOG.info(f"Stack {self.stack_name} does not exist. Nothing to collect.")
